"""Tests to verify the zm module."""

import unittest

from zoneminder import zm


class TestZoneMinder(unittest.TestCase):
    """Tests to verify the ZoneMinder class."""

    def test_build_server_url_no_trailing_slash(self):
        """Verifies that server_url is correct with no trailing slashes."""
        host = "https://zoneminder.com"
        server_path = "zm"
        self.assertEqual(
            "https://zoneminder.com/zm/",
            zm.ZoneMinder._build_server_url(host, server_path),
        )

    def test_build_server_url_with_trailing_slash_path(self):
        """Verifies that server_url is correct with trailing slash in path."""
        host = "https://zoneminder.com"
        server_path = "/zm/"
        self.assertEqual(
            "https://zoneminder.com/zm/",
            zm.ZoneMinder._build_server_url(host, server_path),
        )

    def test_build_server_url_with_trailing_slash_host(self):
        """Verifies that server_url is correct with trailing slash in host."""
        host = "https://zoneminder.com/"
        server_path = "zm/"
        self.assertEqual(
            "https://zoneminder.com/zm/",
            zm.ZoneMinder._build_server_url(host, server_path),
        )

    def test_build_server_url_with_both_trailing_slash(self):
        """Verifies that server_url is correct with trailing slash in both."""
        host = "https://zoneminder.com/"
        server_path = "/zm/"
        self.assertEqual(
            "https://zoneminder.com/zm/",
            zm.ZoneMinder._build_server_url(host, server_path),
        )

    def test_get_zms_url_no_trailing_slash(self):
        """Verifies that zms_url is correct with no trailing slashes."""
        client = zm.ZoneMinder(
            "http://zoneminder.com",
            None,
            None,
            server_path="zm",
            zms_path="cgi-bin/npm-zms",
        )
        self.assertEqual("http://zoneminder.com/cgi-bin/npm-zms", client.get_zms_url())

    def test_get_zms_url_with_leading_slash_zms_path(self):
        """Verifies that zms_url is correct with leading slash in zms_path."""
        client = zm.ZoneMinder(
            "http://zoneminder.com",
            None,
            None,
            server_path="zm",
            zms_path="/cgi-bin/npm-zms",
        )
        self.assertEqual("http://zoneminder.com/cgi-bin/npm-zms", client.get_zms_url())

    def test_get_url_with_auth_username_special(self):
        """Verifies handing of username with special characters is encoded."""
        client = zm.ZoneMinder(None, "@dmin", None, server_path="zm", zms_path="/cgi-bin/npm-zms")
        self.assertEqual(
            "/cgi-bin/npm-zms&user=%40dmin", client.get_url_with_auth(client.get_zms_url())
        )

    def test_get_url_with_auth_password_special(self):
        """Verifies handing of password with special characters is encoded."""
        client = zm.ZoneMinder(
            None, "@dmin", "p@ssword", server_path="zm", zms_path="/cgi-bin/npm-zms"
        )
        self.assertEqual(
            "/cgi-bin/npm-zms&user=%40dmin&pass=p%40ssword",
            client.get_url_with_auth(client.get_zms_url()),
        )



"""Classes that allow interacting with specific ZoneMinder monitors."""

from enum import Enum
import logging
from typing import Optional
from urllib.parse import urlencode

from requests import post

from .exceptions import ControlTypeError, MonitorControlTypeError

_LOGGER = logging.getLogger(__name__)

# From ZoneMinder's web/includes/config.php.in
STATE_ALARM = 3


class ControlType(Enum):
    """Represents the possibles movements types of the Monitor."""

    RIGHT = "moveConRight"
    LEFT = "moveConLeft"
    UP = "moveConUp"
    DOWN = "moveConDown"
    UP_LEFT = "moveConUpLeft"
    UP_RIGHT = "moveConUpRight"
    DOWN_LEFT = "moveConDownLeft"
    DOWN_RIGHT = "moveConDownRight"

    @classmethod
    def from_move(cls, move) -> Enum:
        """Get the corresponding direction from the move.

        Example values: 'right', 'UP-RIGHT', 'down', 'down-left', or 'up_left'.
        """
        for move_key, move_obj in ControlType.__members__.items():
            if move_key == move.upper().replace("-", "_"):
                return move_obj
        raise ControlTypeError()


class MonitorState(Enum):
    """Represents the current state of the Monitor."""

    NONE = "None"
    MONITOR = "Monitor"
    MODECT = "Modect"
    RECORD = "Record"
    MOCORD = "Mocord"
    NODECT = "Nodect"


class TimePeriod(Enum):
    """Represents a period of time to check for events."""

    @property
    def period(self) -> str:
        """Get the period of time."""
        # pylint: disable=unsubscriptable-object
        return self.value[0]

    @property
    def title(self) -> str:
        """Explains what is measured in this period."""
        # pylint: disable=unsubscriptable-object
        return self.value[1]

    @staticmethod
    def get_time_period(value):
        """Get the corresponding TimePeriod from the value.

        Example values: 'all', 'hour', 'day', 'week', or 'month'.
        """
        for time_period in TimePeriod:
            if time_period.period == value:
                return time_period
        raise ValueError(f"{value} is not a valid TimePeriod")

    ALL = ("all", "Events")
    HOUR = ("hour", "Events Last Hour")
    DAY = ("day", "Events Last Day")
    WEEK = ("week", "Events Last Week")
    MONTH = ("month", "Events Last Month")


class Monitor:
    """Represents a Monitor from ZoneMinder."""

    def __init__(self, client, raw_result):
        """Create a new Monitor."""
        self._client = client
        self._raw_result = raw_result
        raw_monitor = raw_result["Monitor"]
        self._monitor_id = int(raw_monitor["Id"])
        self._monitor_url = f"api/monitors/{self._monitor_id}.json"
        self._name = raw_monitor["Name"]
        self._controllable = bool(int(raw_monitor["Controllable"]))
        self._mjpeg_image_url = self._build_image_url(raw_monitor, "jpeg")
        self._still_image_url = self._build_image_url(raw_monitor, "single")
        self._fmt = "{}(id={}, name={}, controllable={})"

    def __repr__(self) -> str:
        """Representation of a Monitor."""
        return self._fmt.format(self.__class__.__name__, self.id, self.name)

    def __str__(self) -> str:
        """Representation of a Monitor."""
        return self.__repr__()

    @property
    def id(self) -> int:
        """Get the ZoneMinder id number of this Monitor."""
        # pylint: disable=invalid-name
        return self._monitor_id

    @property
    def name(self) -> str:
        """Get the name of this Monitor."""
        return self._name

    def update_monitor(self):
        """Update the monitor and monitor status from the ZM server."""
        result = self._client.get_state(self._monitor_url)
        self._raw_result = result["monitor"]

    @property
    def function(self) -> MonitorState:
        """Get the MonitorState of this Monitor."""
        self.update_monitor()

        return MonitorState(self._raw_result["Monitor"]["Function"])

    @function.setter
    def function(self, new_function):
        """Set the MonitorState of this Monitor."""
        self._client.change_state(self._monitor_url, {"Monitor[Function]": new_function.value})

    @property
    def controllable(self) -> bool:
        """Indicate whether this Monitor is movable."""
        return self._controllable

    @property
    def mjpeg_image_url(self) -> str:
        """Get the motion jpeg (mjpeg) image url of this Monitor."""
        return self._mjpeg_image_url

    @property
    def still_image_url(self) -> str:
        """Get the still jpeg image url of this Monitor."""
        return self._still_image_url

    @property
    def is_recording(self) -> Optional[bool]:
        """Indicate if this Monitor is currently recording."""
        status_response = self._client.get_state(
            f"api/monitors/alarm/id:{self._monitor_id}/command:status.json"
        )

        if not status_response:
            _LOGGER.warning("Could not get status for monitor %s.", self._monitor_id)
            return None

        status = status_response.get("status")
        # ZoneMinder API returns an empty string to indicate that this monitor
        # cannot record right now
        try:
            return int(status) == STATE_ALARM
        except ValueError:
            return False

    @property
    def is_available(self) -> bool:
        """Indicate if this Monitor is currently available."""
        status_response = self._client.get_state(
            f"api/monitors/daemonStatus/id:{self._monitor_id}/daemon:zmc.json"
        )

        if not status_response:
            _LOGGER.warning("Could not get availability for monitor %s.", self._monitor_id)
            return False

        # Monitor_Status was only added in ZM 1.32.3
        monitor_status = self._raw_result.get("Monitor_Status", None)
        capture_fps = monitor_status and monitor_status["CaptureFPS"]

        return status_response.get("status", False) and capture_fps != "0.00"

    def get_events(self, time_period, include_archived=False) -> Optional[int]:
        """Get the number of events that have occurred on this Monitor.

        Specifically only gets events that have occurred within the TimePeriod
        provided.
        """
        date_filter = f"1%20{time_period.period}"
        if time_period == TimePeriod.ALL:
            # The consoleEvents API uses DATE_SUB, so give it
            # something large
            date_filter = "100%20year"

        archived_filter = "/Archived=:0"
        if include_archived:
            archived_filter = ""

        event = self._client.get_state(
            f"api/events/consoleEvents/{date_filter}{archived_filter}.json"
        )

        try:
            events_by_monitor = event["results"]
            if isinstance(events_by_monitor, list):
                return 0
            return events_by_monitor.get(str(self._monitor_id), 0)
        except (TypeError, KeyError, AttributeError):
            return None

    def _build_image_url(self, monitor, mode) -> str:
        """Build and return a ZoneMinder camera image url."""
        query = urlencode(
            {
                "mode": mode,
                "buffer": monitor["StreamReplayBuffer"],
                "monitor": monitor["Id"],
            }
        )
        url = f"{self._client.get_zms_url()}?{query}"
        _LOGGER.debug("Monitor %s %s URL (without auth): %s", monitor["Id"], mode, url)
        return self._client.get_url_with_auth(url)

    def ptz_control_command(self, direction, token, base_url) -> bool:
        """Move camera."""
        if not self.controllable:
            raise MonitorControlTypeError()

        ptz_url = f"{base_url}index.php"

        params = {
            "view": "request",
            "request": "control",
            "id": self.id,
            "control": ControlType.from_move(direction).value,
            "xge": 43,
            "token": token,
        }

        req = post(url=ptz_url, params=params, timeout=10)
        return bool(req.ok)



"""An API Client to interact with ZoneMinder."""
import logging
from typing import List, Optional
from urllib.parse import quote, urljoin

import requests

from zoneminder.exceptions import ControlTypeError, MonitorControlTypeError
from zoneminder.monitor import Monitor
from zoneminder.run_state import RunState

_LOGGER = logging.getLogger(__name__)


class ZoneMinder:
    """The ZoneMinder API client itself. Create one of these to begin."""

    DEFAULT_SERVER_PATH = "/zm/"
    DEFAULT_ZMS_PATH = "/zm/cgi-bin/nph-zms"
    DEFAULT_TIMEOUT = 10
    LOGIN_RETRIES = 2
    MONITOR_URL = "api/monitors.json"

    def __init__(
        self,
        server_host,
        username,
        password,
        server_path=DEFAULT_SERVER_PATH,
        zms_path=DEFAULT_ZMS_PATH,
        verify_ssl=True,
    ) -> None:
        """Create a ZoneMinder API Client."""
        self._server_url = ZoneMinder._build_server_url(server_host, server_path)
        self._zms_url = ZoneMinder._build_zms_url(server_host, zms_path)
        self._username = username
        self._password = password
        self._verify_ssl = verify_ssl
        self._cookies = None
        self._auth_token = None

    def login(self):
        """Login to the ZoneMinder API."""
        _LOGGER.debug("Attempting to login to ZoneMinder")

        login_post = {}
        if self._username:
            login_post["user"] = self._username
        if self._password:
            login_post["pass"] = self._password

        req = requests.post(
            urljoin(self._server_url, "api/host/login.json"),
            data=login_post,
            verify=self._verify_ssl,
            timeout=ZoneMinder.DEFAULT_TIMEOUT,
        )
        if req.ok:
            try:
                self._auth_token = req.json()["access_token"]
                return True
            except KeyError:
                # Try legacy auth below
                pass

        return self._legacy_auth()

    def _legacy_auth(self):
        login_post = {"view": "console", "action": "login"}
        if self._username:
            login_post["username"] = self._username
        if self._password:
            login_post["password"] = self._password

        req = requests.post(
            urljoin(self._server_url, "index.php"),
            data=login_post,
            verify=self._verify_ssl,
            timeout=ZoneMinder.DEFAULT_TIMEOUT,
        )
        self._cookies = req.cookies

        # Login calls returns a 200 response on both failure and success.
        # The only way to tell if you logged in correctly is to issue an api
        # call.
        req = requests.get(
            urljoin(self._server_url, "api/host/getVersion.json"),
            cookies=self._cookies,
            timeout=ZoneMinder.DEFAULT_TIMEOUT,
            verify=self._verify_ssl,
        )

        if not req.ok:
            _LOGGER.error("Connection error logging into ZoneMinder")
            return False

        return True

    def get_state(self, api_url) -> dict:
        """Perform a GET request on the specified ZoneMinder API URL."""
        return self._zm_request("get", api_url)

    def change_state(self, api_url, post_data) -> dict:
        """Perform a POST request on the specific ZoneMinder API Url."""
        return self._zm_request("post", api_url, post_data)

    def _zm_request(self, method, api_url, data=None, timeout=DEFAULT_TIMEOUT) -> dict:
        """Perform a request to the ZoneMinder API."""
        token_url_suffix = ""
        if self._auth_token:
            token_url_suffix = "?token=" + self._auth_token
        try:
            # Since the API uses sessions that expire, sometimes we need to
            # re-auth if the call fails.
            for _ in range(ZoneMinder.LOGIN_RETRIES):
                req = requests.request(
                    method,
                    urljoin(self._server_url, api_url) + token_url_suffix,
                    data=data,
                    cookies=self._cookies,
                    timeout=timeout,
                    verify=self._verify_ssl,
                )

                if not req.ok:
                    self.login()
                else:
                    break

            else:
                _LOGGER.error("Unable to get API response from ZoneMinder")

            try:
                return req.json()
            except ValueError:
                _LOGGER.exception(
                    'JSON decode exception caught while attempting to decode "%s"',
                    req.text,
                )
                return {}
        except requests.exceptions.ConnectionError:
            _LOGGER.exception("Unable to connect to ZoneMinder")
            return {}

    def get_monitors(self) -> List[Monitor]:
        """Get a list of Monitors from the ZoneMinder API."""
        raw_monitors = self._zm_request("get", ZoneMinder.MONITOR_URL)
        if not raw_monitors:
            _LOGGER.warning("Could not fetch monitors from ZoneMinder")
            return []

        if "monitors" not in raw_monitors:
            _LOGGER.warning("Could not parse list of monitors from ZoneMinder")
            return []

        monitors = []
        for raw_result in raw_monitors["monitors"]:
            _LOGGER.debug("Initializing camera %s", raw_result["Monitor"]["Id"])
            monitors.append(Monitor(self, raw_result))

        return monitors

    def get_run_states(self) -> List[RunState]:
        """Get a list of RunStates from the ZoneMinder API."""
        raw_states = self.get_state("api/states.json")
        if not raw_states:
            _LOGGER.warning("Could not fetch runstates from ZoneMinder")
            return []

        run_states = []
        for i in raw_states["states"]:
            raw_state = i["State"]
            _LOGGER.info("Initializing runstate %s", raw_state["Id"])
            run_states.append(RunState(self, raw_state))

        return run_states

    def get_active_state(self) -> Optional[str]:
        """Get the name of the active run state from the ZoneMinder API."""
        for state in self.get_run_states():
            if state.active:
                return state.name
        return None

    def set_active_state(self, state_name):
        """
        Set the ZoneMinder run state to the given state name, via ZM API.

        Note that this is a long-running API call; ZoneMinder changes the state
        of each camera in turn, and this GET does not receive a response until
        all cameras have been updated. Even on a reasonably powerful machine,
        this call can take ten (10) or more seconds **per camera**. This method
        sets a timeout of 120, which should be adequate for most users.
        """
        _LOGGER.info("Setting ZoneMinder run state to state %s", state_name)
        return self._zm_request("GET", f"api/states/change/{state_name}.json", timeout=120)

    def get_zms_url(self) -> str:
        """Get the url to the current ZMS instance."""
        return self._zms_url

    def get_url_with_auth(self, url) -> str:
        """Add the auth credentials to a url (if needed)."""
        if not self._username:
            return url

        url += f"&user={quote(self._username)}"

        if not self._password:
            return url
        return url + f"&pass={quote(self._password)}"

    @property
    def is_available(self) -> bool:
        """Indicate if this ZoneMinder service is currently available."""
        status_response = self.get_state("api/host/daemonCheck.json")

        if not status_response:
            return False

        return status_response.get("result") == 1

    @property
    def verify_ssl(self) -> bool:
        """Indicate whether urls with http(s) should verify the certificate."""
        return self._verify_ssl

    @staticmethod
    def _build_zms_url(server_host, zms_path) -> str:
        """Build the ZMS url to the current ZMS instance."""
        return urljoin(server_host, zms_path)

    @staticmethod
    def _build_server_url(server_host, server_path) -> str:
        """Build the server url making sure it ends in a trailing slash."""
        server_url = urljoin(server_host, server_path)
        if server_url[-1] == "/":
            return server_url
        return f"{server_url}/"

    def move_monitor(self, monitor: Monitor, direction: str):
        """Call Zoneminder to move."""
        try:
            result = monitor.ptz_control_command(direction, self._auth_token, self._server_url)
            if result:
                _LOGGER.info("Success to move camera to %s", direction)
            else:
                _LOGGER.error("Impossible to move camera to %s", direction)
        except ControlTypeError:
            _LOGGER.exception("Impossible move monitor")
        except MonitorControlTypeError:
            _LOGGER.exception("Impossible to use direction")



"""Classes that allow interacting with ZoneMinder RunStates."""


class RunState:
    """Represents a Run State from ZoneMinder."""

    def __init__(self, client, raw_state):
        """Create a new RunState."""
        self._client = client
        self._state_id = int(raw_state["Id"])
        self._state_url = "api/states.json"
        self._name = raw_state["Name"]

    @property
    def id(self) -> int:
        """Get the ZoneMinder id number of this RunState."""
        # pylint: disable=invalid-name
        return self._state_id

    @property
    def name(self) -> str:
        """Get the name of this RunState."""
        return self._name

    @property
    def active(self) -> bool:
        """Indicate if this RunState is currently active."""
        states = self._client.get_state(self._state_url)["states"]
        for state in states:
            state = state["State"]
            if int(state["Id"]) == self._state_id:
                # yes, the ZM API uses the *string* "1" for this...
                # Since ZM 1.36 this is now an *int*, provides support for legacy versions
                return int(state["IsActive"]) == 1
        return False

    def activate(self):
        """Activate this RunState."""
        self._client.set_active_state(self._name)



"""This module contains the set of zoneminder exceptions."""


class ZoneminderError(Exception):
    """General Zoneminder error occurred."""

    def __init__(self, value=None):
        """Create a generic exception."""
        super().__init__(value)
        self.value = value

    def __str__(self):
        """Get exception message based in docstring."""
        msg = self.__class__.__doc__
        if self.value is not None:
            msg = msg.rstrip(".")
            msg += ": " + repr(self.value) + "."
        return msg


class CommError(ZoneminderError):
    """A communication error occurred."""


class LoginError(ZoneminderError):
    """A login error occurred."""


class ControlTypeError(ZoneminderError):
    """Unexpected move direction."""


class MonitorControlTypeError(ZoneminderError):
    """Unexpected command to monitor."""


class PtzError(ZoneminderError):
    """A control error occurred."""



"""Tests to verify ControlType."""


from pytest import mark, raises

from zoneminder.exceptions import ControlTypeError
from zoneminder.monitor import ControlType


@mark.parametrize(
    "source,want",
    [
        ("right", ControlType.RIGHT),
        ("LEFT", ControlType.LEFT),
        ("UP_LEFT", ControlType.UP_LEFT),
        ("down-LEFT", ControlType.DOWN_LEFT),
    ],
)
def test_control_type_from_move(source, want):
    """Verifies that ControlType return correct move type."""
    assert ControlType.from_move(source) == want


@mark.parametrize(
    "source,want",
    [
        ("UP-DOWN", raises(ControlTypeError)),
        ("rigth", raises(ControlTypeError)),
    ],
)
def test_control_type_from_move_wrong_move(source, want):
    """Verifies that ControlType return correct exceptions."""
    with want:
        ControlType.from_move(source)
